package spinn3rreaderd;

import com.spinn3r.api.EntryDecoderFactory;
import com.spinn3r.api.protobuf.ContentApi;
import com.spinn3r.api.util.Decoder;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

/**
 * 
 * @author Klemen
 */
public class Main {
  // Sleeping time
  public static final long SLEEP_TIME = 100 * 1000;
  public static final long MAX_FILE_LENGTH = 2L * 1000 * 1000 * 1000;
  // Input arguments for the deamon
  public static File m_inputDir = null;
  public static File m_outputDir = null;
  // Transaction manager
  public static TransactionMng m_trxMng = null;
  // Current transaction
  public static File m_outputFile = null;

  //
  //
  public static String getMostFrequentDate(ArrayList<Entry> entries) {
    /*
     * This function finds the most frequent date (Y:M:D:H) in the given entries.
     */
    // Calculate the frequency for each date
    HashMap<String, Integer> freqs = new HashMap<String, Integer>();
    for (Entry e : entries) {
      // Let's take only the part of the date that consits of Year, Month, Day and Hour
      String date = e.m_pubDate.toString();
      int index = date.indexOf(':');
      String newdate = date.substring(0, index);
      //
      Integer f = freqs.get(newdate);
      if (f == null)
        freqs.put(newdate, 1);
      else
        f++;
    }
    // Find the date with the highest frequency
    String date = null;
    int freq = 0;
    for (Map.Entry<String, Integer> entry : freqs.entrySet()) {
      if (freq < entry.getValue()) {
        freq = entry.getValue();
        date = entry.getKey();
      }
    }
    //
    return date;
  }

  public static String getSuitableFilename(String prefix, String postfix) {
    /*
     * This function finds the suitable filename with given prefix and postfix. Prefix and posfix
     * should be two dates which mark the beginning and the end of the date entries respectively.
     */
    // Compose a new filename
    String filename = prefix + "--" + postfix;
    // Check if this filename already exists in the output directory
    File f = new File(m_outputDir, filename);
    if (!f.exists())
      return filename;
    // We add some integer to the filename
    int i = 0;
    do {
      f = new File(m_outputDir, filename + "_" + String.valueOf(i));
      i++;
    } while (f.exists());
    //
    return filename + "_" + String.valueOf(i - 1);
  }

  public static void parseInput(String[] args) throws Exception {
    /*
     * We have the folowing arguments: (1) Input directory for .protostream files (2) Output
     * directory for parsed and cleaned files (3) Output file name
     */
    if (args.length < 3)
      throw new Exception(
          "Too few input arguments: inputDirectory, outputDirectory, outputFilename");

    // (1)
    String inputDir = args[0];
    m_inputDir = new File(inputDir);
    if (!m_inputDir.isDirectory())
      throw new Exception(String.format("Watch directory path is not a directory: %s", inputDir));

    // (2)
    String outputDir = args[1];
    m_outputDir = new File(outputDir);
    if (!m_outputDir.isDirectory())
      throw new Exception(String.format("Watch directory path is not a directory: %s", outputDir));

    // (3)
    String outputFile = args[2];
    m_outputFile = new File(m_outputDir, outputFile);
  }

  public static void init() throws Exception {
    /*
     * We need to read the last successful transaction from file transaction.txt. Once we have the
     * last transaction, we are ready to the infinite loop of the deamon.
     */

    // Set the transaction manager
    m_trxMng = new TransactionMng(m_outputDir.getPath());
    // Read the transaction
    String filename = m_trxMng.readTransaction();
    // File to the current transaction
    m_outputFile = new File(m_outputDir, filename);
  }

  public static void run() throws Exception {
    // Create protostream file iterator
    ProtostreamFileIt fileIt = new ProtostreamFileIt(m_inputDir.getPath());
    // Add protostream files to the list.
    fileIt.clear();
    fileIt.addFiles();
    // Objects for saving processed data into .txt file
    // FileWriter fw = new FileWriter(m_outputFile, true); //Second argument has to be TRUE
    // BufferedWriter bw = new BufferedWriter(fw);

    BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(
        m_outputFile, true), "UTF-8")); // true: append to file.

    File file = null;
    ArrayList<Entry> entries = new ArrayList<Entry>();

    while ((file = fileIt.next()) != null) {
      // Decode the file
      EntryDecoderFactory factory = EntryDecoderFactory.newFactory();

      // Clean and parse current file
      Decoder<ContentApi.Entry> decoder = factory.get(file);
      ProtostreamParser pp = new ProtostreamParser(decoder);
      pp.parse();
      decoder.close();

      // Get entries
      entries = pp.getEntries();

      // if(entries.isEmpty()) { file.delete(); continue; }

      //
      // Write to disk all the processed entries
      for (Entry e : entries)
        e.write(bw);
      bw.flush();

      // Delete current protostream file
      // file.delete();

      //
      //
      // This "if" section is about RARING the output file, creating a new output file, ...
      // if(m_outputFile.length() > MAX_FILE_LENGTH){
      // ...removed
      // }
    }

    /*
     * No RAR compression (commented out by BOB)
     * 
     * Date now = new Date(); System.out.println(now.toString() + " Compressing to rar ...");
     * 
     * //Creata a new output file, ... bw.close(); fw.close();
     * 
     * //Get the most frequent date in the entries //String postfix = getMostFrequentDate(entries);
     * // //Find the suitable filename for RAR file //String filename =
     * getSuitableFilename(m_outputFile.getName(), postfix);
     * 
     * String filename = m_outputFile.getName(); // Append .rar to the file name filename =
     * filename.substring(0, filename.lastIndexOf(".")) + ".rar";
     * 
     * // Compress .txt to .rar RAR.RARFile(m_outputFile, new File(m_outputDir, filename));
     * 
     * //Write the new output filename //m_trxMng.writeTransaction(postfix);
     * 
     * //Delete .txt file m_outputFile.delete();
     */

    // Prepare everything for writing into a new output file
    // filename = m_trxMng.readTransaction();
    // m_outputFile = new File(m_outputDir, filename);
    // fw = new FileWriter(m_outputFile, true);
    // bw = new BufferedWriter(fw);
    // System.out.println(file.getName());
    // Thread.sleep(SLEEP_TIME);
    // Add any new files
    // fileIt.clear(); fileIt.addFiles();
  }

  public static void main(String[] args) throws Exception {
    Date now = new Date();
    System.out.println(now.toString() + " Starting Spinn3rreaderd ...");

    // Parse input parameters
    now = new Date();
    System.out.println(now.toString() + " Parsing input parameters ...");
    parseInput(args);

    //
    now = new Date();
    System.out.println(now.toString() + " Initializing ...");
    // init();

    //
    now = new Date();
    System.out.println(now.toString() + " Converting protostreams ...");
    run();

    now = new Date();
    System.out.println(now.toString() + " Ending Spinn3rreaderd ...");
  }
}
