/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package spinn3rreaderd;

import java.io.File;
import java.text.ParseException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * @author Klemen
 * Provides file by file iteration in a given directory in ascending order of last modified date. 
 * If there are no files, it return null.
 */
public class ProtostreamFileIt {
    private File m_dir;
    private LinkedList<File> m_files;

    public ProtostreamFileIt(String directoryPath) throws ParseException {
        m_dir = new File(directoryPath);
        if (!m_dir.isDirectory()) throw new ParseException( String.format( "Watch directory path is not a directory: %s", directoryPath ), -1);
        //
        m_files = new LinkedList();
    }
    public boolean hasNext(){
        return m_files.size() > 0;
    }
    public File next() {
        return m_files.pollFirst();
    }
    public void clear(){
        m_files.clear();
    }
    public void addFiles(){ //Warning: clear() should be always called before addFiles(), so that no duplicates appear in the list.
        //Get files from current directory
        File[] files = m_dir.listFiles();
        //Sort files by last modified date
        Arrays.sort(files, new Comparator<File>(){@Override public int compare(File f1, File f2) { return Long.valueOf(f1.lastModified()).compareTo(f2.lastModified()); } });  
        //Store them in a linked list
        m_files.addAll(Arrays.asList(files));
        //Delete files, that are not .protostream files
        ListIterator<File> it = m_files.listIterator();
        while(it.hasNext()){
            File f = it.next();
            if(!f.getName().endsWith("protostream")) it.remove();
        }
    }
}