package spinn3rreaderd;

import com.spinn3r.api.protobuf.ContentApi;
import com.spinn3r.api.util.CompressedBLOB;
import com.spinn3r.api.util.Decoder;

import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.process.PTBTokenizer;
import edu.stanford.nlp.process.WordTokenFactory;
import edu.stanford.nlp.util.StringUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.io.StringReader;
import java.lang.Integer;
import java.net.URLDecoder;

import org.apache.commons.lang.StringEscapeUtils;

/**
 * @author west1 (adapting code by klemen)
 */
public class ProtostreamParser {
	private Decoder<ContentApi.Entry> m_decoder = null;
	private ArrayList<Entry> m_entries = null;

	private static WordTokenFactory tokenFactory = new WordTokenFactory();

	public ProtostreamParser(Decoder<ContentApi.Entry> decoder) {
		m_decoder = decoder;
		m_entries = new ArrayList<Entry>();
	}

	public static String decompressContent(ContentApi.Content content) throws Exception {
		CompressedBLOB content_blob = new CompressedBLOB(content.getData().toByteArray());
		return content_blob.decompress();
	}

	public void parse() throws Exception {
		// Pass through all the entries in the protostream file.
		ContentApi.Entry entry = null;
		int i = 0;
		while (true) {
			try {
				// Decode the entry. In case decoding fails, we throw away the whole .protostream file.
				System.out.println("__entry_1__");
				i = i + 1;
				System.out.println("__count__ " + Integer.toString(i));
				try {
					entry = m_decoder.read();
				} catch (Exception e) {
					break;
				}
				System.out.println("__entry_2__");
				// If entry is null, then we are at the end of the .protostream file.
				if (entry == null)
					break;
				// //////////////////////////// BOB: for debugging, process only limited data:
				// if (i == 100) break;
				System.out.println("__entry_3__");
				//
				// Read the data from entry
				String URL = entry.getPermalinkEntry().getCanonicalLink().getHref();
				String pe_title = entry.getPermalinkEntry().getTitle();
				String fe_title = entry.getFeedEntry().getTitle();
				String pubDate = entry.getPermalinkEntry().getLastPublished();
				if (pubDate == null || pubDate.length() == 0) {
					pubDate = entry.getPermalinkEntry().getDateFound();
				}
				String htmlTitle = null;
				String content = decompressContent(entry.getPermalinkEntry().getContentExtract());
				String htmlContent = null;
				ArrayList<StrPosLenTriple> URLs = new ArrayList<StrPosLenTriple>();
				ArrayList<PosLenPair> quotes = new ArrayList<PosLenPair>();
				System.out.println("__entry_3a__");
				//
				// Find the suitable title
				String title = null;
				if (pe_title == null)
					title = fe_title;
				else if (fe_title == null)
					title = pe_title;
				else if (pe_title.length() >= fe_title.length())
					title = pe_title;
				else
					title = fe_title;
				// remove newlines in title, and remove tabs:
				title = clean(title, false);
				htmlTitle = title;
				//
				// We need to check if all the data is suitable for further processing
				System.out.println("__entry_3a0__");
				if (URL == null || URL.length() == 0)
					continue;
				System.out.println("__entry_3a1__");
				if (title == null || title.length() == 0)
					continue;
				System.out.println("__entry_3a2__");
				if (pubDate == null || pubDate.length() == 0)
					continue;
				System.out.println("__entry_3b__");
				//
				// We have all the "canonical" data that we need to process it.
				//
				// Clean URL; escaping newlines and tabs via cleanURL() is important because some URL seem to
				// contain such chars:
				URL = escapeTabsAndNewlines(URLDecoder.decode(URL.trim(), "UTF-8"));
				if (URL.isEmpty())
					continue;
				System.out.println("__entry_3c__");
				//
				// Clean title as much as possible before we extract URLs
				// Extract URLs from title
				title = StringUtils.join(tokenize(title, false, URLs, null), " ");
				//
				// escape date:
				pubDate = escapeTabsAndNewlines(pubDate);
				// important: the URLs extracted from the title will have associated with them pos and len
				// referring to indices in the title string; however, we want pos and len of URLs to refer to
				// the plain-text content string. So we have to let pos and len of URLs found in the title be
				// undefined, encoded as -1:
				for (StrPosLenTriple url : URLs) {
					url.pos = -1;
					url.len = -1;
				}
				if (title.isEmpty())
					continue;
				//
				// We can already store some of the cleaned and processed data into Entry object.
				Entry e = new Entry();
				e.m_URL = URL;
				e.m_title = title;
				e.m_htmlTitle = htmlTitle;
				e.m_pubDate = pubDate;
				//
				// content
				if (content != null) {
					// fork a raw version of the String, including HTML tags and without any further
					// processing; we only make sure that there no tabs and that line breaks are represented
					// by "*NL*", so everything fits on one line in the output:
					htmlContent = clean(content, true);
					// delete newline chars and tabs:
					content = clean(content, false);
					// clean the plain-text content by whitespace-concatenating the tokens; along the way,
					// URLs and quotes are collected:
					content = StringUtils.join(tokenize(content, true, URLs, quotes), " ");
				}
				//
				System.out.println("__entry_4__");
				// Store any possible URLs from title and/or content:
				if (URLs.size() > 0)
					e.m_URLs = URLs;
				if (content != null && !content.isEmpty()) {
					System.out.println("__entry_5__");
					e.m_content = content;
					e.m_htmlContent = htmlContent;
					if (quotes.size() > 0)
						e.m_quotes = quotes;
				}
				// Store the whole entry.
				m_entries.add(e);
			} catch (Exception e) {
				// if an exception occurred, report it on standard out and continue with the next item in the
				// protostream file:
				System.out.println("__error__" + e.getMessage());
				e.printStackTrace(System.out);
			}
		}
	}

	public ArrayList<Entry> getEntries() {
		return m_entries;
	}

	// adapted from
	// http://stackoverflow.com/questions/4742183/java-regex-using-functions-methos-in-string-replaceall
	// (currently unused)
	private static String removeNewlinesFromHTMLTags(String input) {
		StringBuffer output = new StringBuffer();
		Matcher matcher = TAG_PATTERN.matcher(input);
		while (matcher.find()) {
			matcher.appendReplacement(output, matcher.group().replaceAll("(\\r|\\n|\\f)+", " "));
		}
		matcher.appendTail(output);
		return output.toString();
	}

	// if keepNewLines==true, line breaks are kept, but they're now represented by "*NL*"; multiple
	// consecutive line breaks are represented as a single "*NL*" by this method:
	private static String clean(String str, boolean keepNewLines) {
		// remove leading and trailing whitespaces:
		str = str.trim();
		// replace remaining newline characters:
		str = str.replaceAll("(\\r|\\n|\\f)+", keepNewLines ? "\\\\n" : " ");
		// make sure there are no tabs, since those will be used as delimiters in the output:
		str = str.replaceAll("\t+", " ");
		return str;
	}

	// simply escape newlines and tabs:
	private static String escapeTabsAndNewlines(String url) {
		url = url.replace("\n", "\\n");
		url = url.replace("\r", "\\r");
		url = url.replace("\f", "\\f");
		url = url.replace("\t", "\\t");
		return url;
	}

	private static final Pattern TAG_PATTERN = Pattern.compile("(?s)<[^>]*>");
	private static final Pattern HREF_PATTERN = Pattern.compile("(?i)<a.*href=[\\\"'](https?://.*?)[\\\"'].*>");

	// if you don't want to extract urls and/or quotes, set the respective argument to null:
	private static List<String> tokenize(String line, boolean keepNewLines, List<StrPosLenTriple> urls, List<PosLenPair> quotes) {
		// if keepLines==true, newlines are converted to "*NL*", otherwise they're treated as whitespace:
		// PTBTokenizer<Word> tokenizer = PTBTokenizer.newPTBTokenizer(new StringReader(line), keepNewLines);
		String optionString = "tokenizeNLs=%b,americanize=false,normalizeCurrency=false,normalizeParentheses=false,"
				+ "normalizeOtherBrackets=false,unicodeQuotes=false,ptb3Ellipsis=true,escapeForwardSlashAsterisk=false,untokenizable=noneKeep";
		PTBTokenizer<Word> tokenizer = new PTBTokenizer<Word>(new StringReader(line), tokenFactory, String.format(optionString, keepNewLines));
		List<String> words = new ArrayList<String>();
		// stores the index to process next, in the string that would result from concatenating all tokens,
		// separated by whitespace:
		int curIdx = 0;
		// index of last seen opening quote (counting done as for curIdx, where the index refers to the first
		// letter of the quote text, without quotation marks):
		int openSingleQuoteIdx = -1;
		int openDoubleQuoteIdx = -1;
		// same for hrefs:
		int openHrefIdx = -1;
		// always remember the last seen URL:
		String lastURL = null;
		while (tokenizer.hasNext()) {
			String token = tokenizer.next().word();
			Matcher hrefMatcher = HREF_PATTERN.matcher(token);
			// we found a hyperlink:
			if (hrefMatcher.matches()) {
				if (urls != null) {
					openHrefIdx = curIdx;
					lastURL = hrefMatcher.group(1);
				}
			} else if (token.equals("</a>") && openHrefIdx >= 0) {
				if (urls != null) {
					// Remember this hyperlink:
					// Sometimes we encounter empty link spans: <a href=...> </a>; these are a corner case
					// that needs to be handled separately, for the following reason: tokenization separates
					// HTML tags from the surrounding text by one whitespace, so a link span of length L = 0
					// has 1 = L + 1 characters between the opening and closing <a> tags, whereas a link span
					// of length L > 0 has L + 2 characters in between (one whitespace to the left, and one to
					// the right):
					int len = Math.max(0, curIdx - 1 - openHrefIdx);
					urls.add(new StrPosLenTriple(lastURL, openHrefIdx, len));
					// reset the finite state automaton:
					openHrefIdx = -1;
					lastURL = null;
				}
			}
			// ignore this token if it's a non-hyperlink HTML tag:
			else if (!(token.startsWith("<") && token.endsWith(">"))) {
				// now that HTML has been discarded, we don't need '/' to be escaped anymore:
				token = token.replace("\\/", "/");
				// HTML decode, e.g., '&gt;' => '>', '&uuml;' => 'ü':
				token = StringEscapeUtils.unescapeHtml(token);
				// Is this token a URL (plain URL in text, not in href tag)?
				if (urls != null && (token.startsWith("http://") || token.startsWith("https://"))) {
					urls.add(new StrPosLenTriple(token, curIdx, token.length()));
				}
				// Perform one step in the finite state automaton for recognizing quotes:
				if (quotes != null) {
					// double quotes:
					if (token.equals("``")) {
						// reset the opening; +3 is because we want to point to right after the quotation
						// marks:
						openDoubleQuoteIdx = curIdx + 3;
					} else if (token.equals("\'\'") && openDoubleQuoteIdx >= 0) {
						// this to handle empty quotes, where the same issue arises as for empty <a> spans,
						// cf. there:
						int len = Math.max(0, curIdx - 1 - openDoubleQuoteIdx);
						// remember this quote:
						quotes.add(new PosLenPair(openDoubleQuoteIdx, len));
						// reset the finite state automaton:
						openDoubleQuoteIdx = -1;
					}
					// single quotes:
					else if (token.equals("`")) {
						// reset the opening; +2 is because we want to point to right after the quotation
						// marks:
						openSingleQuoteIdx = curIdx + 2;
					} else if (token.equals("\'") && openSingleQuoteIdx >= 0) {
						// this to handle empty quotes, where the same issue arises as for empty <a> spans,
						// cf. there:
						int len = Math.max(0, curIdx - 1 - openSingleQuoteIdx);
						// remember this quote:
						quotes.add(new PosLenPair(openSingleQuoteIdx, len));
						// reset the finite state automaton:
						openSingleQuoteIdx = -1;
					}
				}
				// remember this token:
				words.add(token);
				// advance the index pointer:
				curIdx += token.length() + 1;
			}
		}
		return words;
	}

	public static void main(String[] args) {
		String str = "\"test quote.\" <a href=\"http://www.searchInput.de\">Suche</a>  <p>are offered single as tall as 30%</p>";
		List<StrPosLenTriple> urls = new ArrayList<StrPosLenTriple>();
		List<PosLenPair> quotes = new ArrayList<PosLenPair>();
		System.out.println(edu.stanford.nlp.util.StringUtils.join(tokenize(str, true, urls, quotes)));
		System.out.println(urls);
		System.out.println(quotes);
	}
}
